@@ -26,17 +26,17 @@
  *
  * Encodes raw video streams into VP9 bitstreams.
  *
- * <refsect2>
- * <title>Example launch line</title>
+ * ## Example launch line
+ *
  * |[
  *  gst-launch-1.0 -ev videotestsrc num-buffers=60 ! timeoverlay ! vaapivp9enc ! matroskamux ! filesink location=test.mkv
  * ]|
- * </refsect2>
  */
 
 #include "gstcompat.h"
 #include <gst/vaapi/gstvaapidisplay.h>
 #include <gst/vaapi/gstvaapiencoder_vp9.h>
+#include <gst/vaapi/gstvaapiutils_vpx.h>
 #include "gstvaapiencode_vp9.h"
 #include "gstvaapipluginutil.h"
 #include "gstvaapivideomemory.h"
@@ -50,42 +50,16 @@ GST_DEBUG_CATEGORY_STATIC (gst_vaapi_vp9_encode_debug);
 #define GST_CODEC_CAPS                          \
   "video/x-vp9"
 
-/* *INDENT-OFF* */
-static const char gst_vaapiencode_vp9_sink_caps_str[] =
-  GST_VAAPI_MAKE_SURFACE_CAPS ", "
-  GST_CAPS_INTERLACED_FALSE "; "
-  GST_VIDEO_CAPS_MAKE (GST_VIDEO_FORMATS_ALL) ", "
-  GST_CAPS_INTERLACED_FALSE;
-/* *INDENT-ON* */
-
-/* *INDENT-OFF* */
-static const char gst_vaapiencode_vp9_src_caps_str[] =
-  GST_CODEC_CAPS;
-/* *INDENT-ON* */
-
-/* *INDENT-OFF* */
-static GstStaticPadTemplate gst_vaapiencode_vp9_sink_factory =
-  GST_STATIC_PAD_TEMPLATE ("sink",
-      GST_PAD_SINK,
-      GST_PAD_ALWAYS,
-      GST_STATIC_CAPS (gst_vaapiencode_vp9_sink_caps_str));
-/* *INDENT-ON* */
-
-/* *INDENT-OFF* */
-static GstStaticPadTemplate gst_vaapiencode_vp9_src_factory =
-  GST_STATIC_PAD_TEMPLATE ("src",
-      GST_PAD_SRC,
-      GST_PAD_ALWAYS,
-      GST_STATIC_CAPS (gst_vaapiencode_vp9_src_caps_str));
-/* *INDENT-ON* */
+#define EXTRA_FORMATS {}
 
 /* vp9 encode */
-G_DEFINE_TYPE (GstVaapiEncodeVP9, gst_vaapiencode_vp9, GST_TYPE_VAAPIENCODE);
+GST_VAAPI_ENCODE_REGISTER_TYPE (vp9, VP9, VP9, EXTRA_FORMATS,
+    gst_vaapi_utils_vp9_get_profile_string);
 
 static void
 gst_vaapiencode_vp9_init (GstVaapiEncodeVP9 * encode)
 {
-  gst_vaapiencode_init_properties (GST_VAAPIENCODE_CAST (encode));
+  /* nothing to do here */
 }
 
 static void
@@ -94,34 +68,12 @@ gst_vaapiencode_vp9_finalize (GObject * object)
   G_OBJECT_CLASS (gst_vaapiencode_vp9_parent_class)->finalize (object);
 }
 
-static void
-gst_vaapiencode_vp9_set_property (GObject * object,
-    guint prop_id, const GValue * value, GParamSpec * pspec)
+static GArray *
+gst_vaapiencode_vp9_get_allowed_profiles (GstVaapiEncode * encode,
+    GstCaps * allowed)
 {
-  GstVaapiEncodeClass *const encode_class = GST_VAAPIENCODE_GET_CLASS (object);
-  GstVaapiEncode *const base_encode = GST_VAAPIENCODE_CAST (object);
-
-  switch (prop_id) {
-    default:
-      if (!encode_class->set_property (base_encode, prop_id, value))
-        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-      break;
-  }
-}
-
-static void
-gst_vaapiencode_vp9_get_property (GObject * object,
-    guint prop_id, GValue * value, GParamSpec * pspec)
-{
-  GstVaapiEncodeClass *const encode_class = GST_VAAPIENCODE_GET_CLASS (object);
-  GstVaapiEncode *const base_encode = GST_VAAPIENCODE_CAST (object);
-
-  switch (prop_id) {
-    default:
-      if (!encode_class->get_property (base_encode, prop_id, value))
-        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-      break;
-  }
+  return gst_vaapi_encoder_get_profiles_from_caps (allowed,
+      gst_vaapi_utils_vp9_get_profile_from_string);
 }
 
 static GstCaps *
@@ -142,20 +94,22 @@ gst_vaapiencode_vp9_alloc_encoder (GstVaapiEncode * base,
 }
 
 static void
-gst_vaapiencode_vp9_class_init (GstVaapiEncodeVP9Class * klass)
+gst_vaapiencode_vp9_class_init (GstVaapiEncodeVP9Class * klass, gpointer data)
 {
   GObjectClass *const object_class = G_OBJECT_CLASS (klass);
   GstElementClass *const element_class = GST_ELEMENT_CLASS (klass);
   GstVaapiEncodeClass *const encode_class = GST_VAAPIENCODE_CLASS (klass);
-
-  GST_DEBUG_CATEGORY_INIT (gst_vaapi_vp9_encode_debug,
-      GST_PLUGIN_NAME, 0, GST_PLUGIN_DESC);
+  GstCaps *sink_caps = ((GstVaapiEncodeInitData *) data)->sink_caps;
+  GstCaps *src_caps = ((GstVaapiEncodeInitData *) data)->src_caps;
+  GstPadTemplate *templ;
+  GstCaps *static_caps;
+  gpointer encoder_class;
 
   object_class->finalize = gst_vaapiencode_vp9_finalize;
-  object_class->set_property = gst_vaapiencode_vp9_set_property;
-  object_class->get_property = gst_vaapiencode_vp9_get_property;
+  object_class->set_property = gst_vaapiencode_set_property_subclass;
+  object_class->get_property = gst_vaapiencode_get_property_subclass;
 
-  encode_class->get_properties = gst_vaapi_encoder_vp9_get_default_properties;
+  encode_class->get_allowed_profiles = gst_vaapiencode_vp9_get_allowed_profiles;
   encode_class->get_caps = gst_vaapiencode_vp9_get_caps;
   encode_class->alloc_encoder = gst_vaapiencode_vp9_alloc_encoder;
 
@@ -166,12 +120,26 @@ gst_vaapiencode_vp9_class_init (GstVaapiEncodeVP9Class * klass)
       "Sreerenj Balachandran <sreerenj.balachandran@intel.com>");
 
   /* sink pad */
-  gst_element_class_add_pad_template (element_class,
-      gst_static_pad_template_get (&gst_vaapiencode_vp9_sink_factory));
+  g_assert (sink_caps);
+  static_caps = gst_caps_from_string (GST_VAAPI_ENCODE_STATIC_SINK_CAPS);
+  templ =
+      gst_pad_template_new ("sink", GST_PAD_SINK, GST_PAD_ALWAYS, sink_caps);
+  gst_pad_template_set_documentation_caps (templ, static_caps);
+  gst_element_class_add_pad_template (element_class, templ);
+  gst_caps_unref (static_caps);
+  gst_caps_unref (sink_caps);
 
   /* src pad */
-  gst_element_class_add_pad_template (element_class,
-      gst_static_pad_template_get (&gst_vaapiencode_vp9_src_factory));
+  g_assert (src_caps);
+  static_caps = gst_caps_from_string (GST_CODEC_CAPS);
+  templ = gst_pad_template_new ("src", GST_PAD_SRC, GST_PAD_ALWAYS, src_caps);
+  gst_pad_template_set_documentation_caps (templ, static_caps);
+  gst_element_class_add_pad_template (element_class, templ);
+  gst_caps_unref (static_caps);
+  gst_caps_unref (src_caps);
 
-  gst_vaapiencode_class_init_properties (encode_class);
+  encoder_class = g_type_class_ref (GST_TYPE_VAAPI_ENCODER_VP9);
+  g_assert (encoder_class);
+  gst_vaapiencode_class_install_properties (encode_class, encoder_class);
+  g_type_class_unref (encoder_class);
 }
