@@ -40,6 +40,43 @@ GST_DEBUG_CATEGORY_STATIC (CAT_PERFORMANCE);
 
 #define BUFFER_POOL_SINK_MIN_BUFFERS 2
 
+#define GST_VAAPI_PAD_PRIVATE(pad) \
+  (GST_VAAPI_PLUGIN_BASE_GET_CLASS(plugin)->get_vaapi_pad_private(plugin, pad))
+
+GstVaapiPadPrivate *
+gst_vaapi_pad_private_new (void)
+{
+  GstVaapiPadPrivate *priv = g_new0 (GstVaapiPadPrivate, 1);
+
+  gst_video_info_init (&priv->info);
+
+  return priv;
+}
+
+void
+gst_vaapi_pad_private_reset (GstVaapiPadPrivate * priv)
+{
+  g_assert (priv);
+
+  gst_caps_replace (&priv->caps, NULL);
+  gst_video_info_init (&priv->info);
+
+  g_clear_object (&priv->buffer_pool);
+  g_clear_object (&priv->allocator);
+
+  priv->buffer_size = 0;
+  priv->caps_is_raw = FALSE;
+
+  g_clear_object (&priv->other_allocator);
+}
+
+void
+gst_vaapi_pad_private_finalize (GstVaapiPadPrivate * priv)
+{
+  gst_vaapi_pad_private_reset (priv);
+  g_free (priv);
+}
+
 /* GstVideoContext interface */
 static void
 plugin_set_display (GstVaapiPluginBase * plugin, GstVaapiDisplay * display)
@@ -118,14 +155,15 @@ _set_cached_surface (GstBuffer * buf, GstVaapiSurface * surface)
 {
   return gst_mini_object_set_qdata (GST_MINI_OBJECT (buf),
       g_quark_from_static_string ("GstVaapiDMABufSurface"), surface,
-      (GDestroyNotify) gst_vaapi_object_unref);
+      (GDestroyNotify) gst_vaapi_surface_unref);
 }
 
 static gboolean
 plugin_update_sinkpad_info_from_buffer (GstVaapiPluginBase * plugin,
-    GstBuffer * buf)
+    GstPad * sinkpad, GstBuffer * buf)
 {
-  GstVideoInfo *const vip = &plugin->sinkpad_info;
+  GstVaapiPadPrivate *sinkpriv = GST_VAAPI_PAD_PRIVATE (sinkpad);
+  GstVideoInfo *const vip = &sinkpriv->info;
   GstVideoMeta *vmeta;
   guint i;
 
@@ -162,10 +200,11 @@ is_dma_buffer (GstBuffer * buf)
 }
 
 static gboolean
-plugin_bind_dma_to_vaapi_buffer (GstVaapiPluginBase * plugin,
+plugin_bind_dma_to_vaapi_buffer (GstVaapiPluginBase * plugin, GstPad * sinkpad,
     GstBuffer * inbuf, GstBuffer * outbuf)
 {
-  GstVideoInfo *const vip = &plugin->sinkpad_info;
+  GstVaapiPadPrivate *sinkpriv = GST_VAAPI_PAD_PRIVATE (sinkpad);
+  GstVideoInfo *const vip = &sinkpriv->info;
   GstVaapiVideoMeta *meta;
   GstVaapiSurface *surface;
   GstVaapiSurfaceProxy *proxy;
@@ -175,7 +214,7 @@ plugin_bind_dma_to_vaapi_buffer (GstVaapiPluginBase * plugin,
   if (fd < 0)
     return FALSE;
 
-  if (!plugin_update_sinkpad_info_from_buffer (plugin, inbuf))
+  if (!plugin_update_sinkpad_info_from_buffer (plugin, sinkpad, inbuf))
     goto error_update_sinkpad_info;
 
   meta = gst_buffer_get_vaapi_video_meta (outbuf);
@@ -228,11 +267,22 @@ plugin_reset_texture_map (GstVaapiPluginBase * plugin)
     gst_vaapi_display_reset_texture_map (plugin->display);
 }
 
+static GstVaapiPadPrivate *
+default_get_vaapi_pad_private (GstVaapiPluginBase * plugin, GstPad * pad)
+{
+  if (plugin->sinkpad == pad)
+    return plugin->sinkpriv;
+
+  g_assert (plugin->srcpad == pad);
+  return plugin->srcpriv;
+}
+
 void
 gst_vaapi_plugin_base_class_init (GstVaapiPluginBaseClass * klass)
 {
   klass->has_interface = default_has_interface;
   klass->display_changed = default_display_changed;
+  klass->get_vaapi_pad_private = default_get_vaapi_pad_private;
 }
 
 void
@@ -245,12 +295,16 @@ gst_vaapi_plugin_base_init (GstVaapiPluginBase * plugin,
 
   /* sink pad */
   plugin->sinkpad = gst_element_get_static_pad (GST_ELEMENT (plugin), "sink");
-  gst_video_info_init (&plugin->sinkpad_info);
+
+  if (plugin->sinkpad)
+    plugin->sinkpriv = gst_vaapi_pad_private_new ();
 
   /* src pad */
   if (!(GST_OBJECT_FLAGS (plugin) & GST_ELEMENT_FLAG_SINK))
     plugin->srcpad = gst_element_get_static_pad (GST_ELEMENT (plugin), "src");
-  gst_video_info_init (&plugin->srcpad_info);
+
+  if (plugin->srcpad)
+    plugin->srcpriv = gst_vaapi_pad_private_new ();
 
   plugin->enable_direct_rendering =
       (g_getenv ("GST_VAAPI_ENABLE_DIRECT_RENDERING") != NULL);
@@ -260,7 +314,14 @@ void
 gst_vaapi_plugin_base_finalize (GstVaapiPluginBase * plugin)
 {
   gst_vaapi_plugin_base_close (plugin);
+  gst_vaapi_display_replace (&plugin->display, NULL);
   g_free (plugin->display_name);
+
+  if (plugin->sinkpriv)
+    gst_vaapi_pad_private_finalize (plugin->sinkpriv);
+  if (plugin->srcpriv)
+    gst_vaapi_pad_private_finalize (plugin->srcpriv);
+
   if (plugin->sinkpad)
     gst_object_unref (plugin->sinkpad);
   if (plugin->srcpad)
@@ -279,6 +340,7 @@ gst_vaapi_plugin_base_finalize (GstVaapiPluginBase * plugin)
 gboolean
 gst_vaapi_plugin_base_open (GstVaapiPluginBase * plugin)
 {
+  gst_caps_replace (&plugin->allowed_raw_caps, NULL);
   return TRUE;
 }
 
@@ -295,24 +357,16 @@ gst_vaapi_plugin_base_close (GstVaapiPluginBase * plugin)
   /* Release vaapi textures first if exist, which refs display object */
   plugin_reset_texture_map (plugin);
 
-  gst_vaapi_display_replace (&plugin->display, NULL);
   gst_object_replace (&plugin->gl_context, NULL);
   gst_object_replace (&plugin->gl_display, NULL);
   gst_object_replace (&plugin->gl_other_context, NULL);
 
-  gst_caps_replace (&plugin->sinkpad_caps, NULL);
-  gst_video_info_init (&plugin->sinkpad_info);
-
-  g_clear_object (&plugin->sinkpad_buffer_pool);
-  g_clear_object (&plugin->srcpad_buffer_pool);
-
-  g_clear_object (&plugin->sinkpad_allocator);
-  g_clear_object (&plugin->srcpad_allocator);
-  g_clear_object (&plugin->other_srcpad_allocator);
-
-  gst_caps_replace (&plugin->srcpad_caps, NULL);
-  gst_video_info_init (&plugin->srcpad_info);
   gst_caps_replace (&plugin->allowed_raw_caps, NULL);
+
+  if (plugin->sinkpriv)
+    gst_vaapi_pad_private_reset (plugin->sinkpriv);
+  if (plugin->srcpriv)
+    gst_vaapi_pad_private_reset (plugin->srcpriv);
 }
 
 /**
@@ -435,9 +489,10 @@ reset_allocator (GstAllocator * allocator, GstVideoInfo * vinfo)
 }
 
 static gboolean
-ensure_sinkpad_allocator (GstVaapiPluginBase * plugin, GstCaps * caps,
-    guint * size)
+ensure_sinkpad_allocator (GstVaapiPluginBase * plugin, GstPad * sinkpad,
+    GstCaps * caps, guint * size)
 {
+  GstVaapiPadPrivate *sinkpriv = GST_VAAPI_PAD_PRIVATE (sinkpad);
   GstVideoInfo vinfo;
   const GstVideoInfo *image_info;
   GstVaapiImageUsageFlags usage_flag =
@@ -446,7 +501,7 @@ ensure_sinkpad_allocator (GstVaapiPluginBase * plugin, GstCaps * caps,
   if (!gst_video_info_from_caps (&vinfo, caps))
     goto error_invalid_caps;
 
-  if (!reset_allocator (plugin->sinkpad_allocator, &vinfo))
+  if (!reset_allocator (sinkpriv->allocator, &vinfo))
     goto bail;
 
   /* enable direct upload if upstream requests raw video */
@@ -454,15 +509,14 @@ ensure_sinkpad_allocator (GstVaapiPluginBase * plugin, GstCaps * caps,
     usage_flag = GST_VAAPI_IMAGE_USAGE_FLAG_DIRECT_UPLOAD;
     GST_INFO_OBJECT (plugin, "enabling direct upload in sink allocator");
   }
-  plugin->sinkpad_allocator =
+  sinkpriv->allocator =
       gst_vaapi_video_allocator_new (plugin->display, &vinfo, 0, usage_flag);
 
 bail:
-  if (!plugin->sinkpad_allocator)
+  if (!sinkpriv->allocator)
     goto error_create_allocator;
 
-  image_info =
-      gst_allocator_get_vaapi_video_info (plugin->sinkpad_allocator, NULL);
+  image_info = gst_allocator_get_vaapi_video_info (sinkpriv->allocator, NULL);
   g_assert (image_info);        /* allocator ought set its image info */
 
   /* update the size with the one generated by the allocator */
@@ -529,29 +583,24 @@ create_dmabuf_srcpad_allocator (GstVaapiPluginBase * plugin,
 }
 
 static gboolean
-ensure_srcpad_allocator (GstVaapiPluginBase * plugin, GstVideoInfo * vinfo,
-    GstCaps * caps)
+ensure_srcpad_allocator (GstVaapiPluginBase * plugin, GstPad * srcpad,
+    GstVideoInfo * vinfo, GstCaps * caps)
 {
-  gboolean different_caps;
+  GstVaapiPadPrivate *srcpriv = GST_VAAPI_PAD_PRIVATE (srcpad);
   const GstVideoInfo *image_info;
 
-  if (!reset_allocator (plugin->srcpad_allocator, vinfo))
+  if (!reset_allocator (srcpriv->allocator, vinfo))
     goto valid_allocator;
 
-  plugin->srcpad_allocator = NULL;
-  if (caps && gst_caps_is_video_raw (caps)) {
-    GstAllocator *allocator = create_dmabuf_srcpad_allocator (plugin, vinfo,
-        !plugin->srcpad_can_dmabuf);
-    plugin->srcpad_allocator = allocator;
-  } else if (caps && gst_vaapi_caps_feature_contains (caps,
+  srcpriv->allocator = NULL;
+  if (caps && gst_vaapi_caps_feature_contains (caps,
           GST_VAAPI_CAPS_FEATURE_DMABUF)) {
-    plugin->srcpad_allocator =
-        create_dmabuf_srcpad_allocator (plugin, vinfo, FALSE);
-    if (!plugin->srcpad_allocator)
+    srcpriv->allocator = create_dmabuf_srcpad_allocator (plugin, vinfo, FALSE);
+    if (!srcpriv->allocator)
       goto error_create_allocator;
   }
 
-  if (!plugin->srcpad_allocator) {
+  if (!srcpriv->allocator) {
     GstVaapiImageUsageFlags usage_flag =
         GST_VAAPI_IMAGE_USAGE_FLAG_NATIVE_FORMATS;
 
@@ -560,46 +609,49 @@ ensure_srcpad_allocator (GstVaapiPluginBase * plugin, GstVideoInfo * vinfo,
       GST_INFO_OBJECT (plugin, "enabling direct rendering in source allocator");
     }
 
-    plugin->srcpad_allocator =
+    srcpriv->allocator =
         gst_vaapi_video_allocator_new (plugin->display, vinfo, 0, usage_flag);
   }
 
-  if (!plugin->srcpad_allocator)
+  if (!srcpriv->allocator)
     goto error_create_allocator;
 
 valid_allocator:
-  image_info =
-      gst_allocator_get_vaapi_video_info (plugin->srcpad_allocator, NULL);
+  image_info = gst_allocator_get_vaapi_video_info (srcpriv->allocator, NULL);
   g_assert (image_info);        /* both allocators ought set its image
                                  * info */
 
   /* update the size with the one generated by the allocator */
   GST_VIDEO_INFO_SIZE (vinfo) = GST_VIDEO_INFO_SIZE (image_info);
 
-  /* the received caps are the "allocation caps" which may be
-   * different from the "negotiation caps". In this case, we should
-   * indicate the allocator to store the negotiation caps since they
-   * are the one should be used for frame mapping with GstVideoMeta */
-  different_caps = GST_IS_VIDEO_DECODER (plugin) && plugin->srcpad_caps &&
-      !gst_caps_is_strictly_equal (plugin->srcpad_caps, caps);
+  if (GST_IS_VIDEO_DECODER (plugin)) {
+    /* the received caps are the "allocation caps" which may be
+     * different from the "negotiation caps". In this case, we should
+     * indicate the allocator to store the negotiation caps since they
+     * are the one should be used for frame mapping with GstVideoMeta */
+    gboolean different_caps = srcpriv->caps &&
+        !gst_caps_is_strictly_equal (srcpriv->caps, caps);
+    const GstVideoInfo *previous_negotiated =
+        gst_allocator_get_vaapi_negotiated_video_info (srcpriv->allocator);
 
-  if (different_caps) {
-    guint i;
-    GstVideoInfo vi = plugin->srcpad_info;
+    if (different_caps) {
+      guint i;
+      GstVideoInfo vi = srcpriv->info;
 
-    /* update the planes and the size with the allocator image/surface
-     * info, but not the resolution */
-    for (i = 0; i < GST_VIDEO_INFO_N_PLANES (image_info); i++) {
-      GST_VIDEO_INFO_PLANE_OFFSET (&vi, i) =
-          GST_VIDEO_INFO_PLANE_OFFSET (image_info, i);
-      GST_VIDEO_INFO_PLANE_STRIDE (&vi, i) =
-          GST_VIDEO_INFO_PLANE_STRIDE (image_info, i);
+      /* update the planes and the size with the allocator image/surface
+       * info, but not the resolution */
+      for (i = 0; i < GST_VIDEO_INFO_N_PLANES (image_info); i++) {
+        GST_VIDEO_INFO_PLANE_OFFSET (&vi, i) =
+            GST_VIDEO_INFO_PLANE_OFFSET (image_info, i);
+        GST_VIDEO_INFO_PLANE_STRIDE (&vi, i) =
+            GST_VIDEO_INFO_PLANE_STRIDE (image_info, i);
+      }
+      GST_VIDEO_INFO_SIZE (&vi) = GST_VIDEO_INFO_SIZE (image_info);
+      gst_allocator_set_vaapi_negotiated_video_info (srcpriv->allocator, &vi);
+    } else if (previous_negotiated) {
+      gst_allocator_set_vaapi_negotiated_video_info (srcpriv->allocator, NULL);
     }
-    GST_VIDEO_INFO_SIZE (&vi) = GST_VIDEO_INFO_SIZE (image_info);
-    gst_allocator_set_vaapi_negotiated_video_info (plugin->srcpad_allocator,
-        &vi);
   }
-
   return TRUE;
 
   /* ERRORS */
@@ -687,16 +739,18 @@ error_pool_config:
 /**
  * ensure_sinkpad_buffer_pool:
  * @plugin: a #GstVaapiPluginBase
- * @caps: the initial #GstCaps for the resulting buffer pool
+ * @sinkpad: the #GstPad to use for the resulting buffer pool
  *
  * Makes sure the sink pad video buffer pool is created with the
- * appropriate @caps.
+ * appropriate caps defined in the @sinkpad.
  *
  * Returns: %TRUE if successful, %FALSE otherwise.
  */
 static gboolean
-ensure_sinkpad_buffer_pool (GstVaapiPluginBase * plugin, GstCaps * caps)
+ensure_sinkpad_buffer_pool (GstVaapiPluginBase * plugin, GstPad * sinkpad)
 {
+  GstVaapiPadPrivate *sinkpriv = GST_VAAPI_PAD_PRIVATE (sinkpad);
+  GstCaps *caps = sinkpriv->caps;
   GstBufferPool *pool;
   guint size;
 
@@ -707,30 +761,104 @@ ensure_sinkpad_buffer_pool (GstVaapiPluginBase * plugin, GstCaps * caps)
   if (!gst_vaapi_plugin_base_ensure_display (plugin))
     return FALSE;
 
-  if (plugin->sinkpad_buffer_pool) {
-    if (gst_vaapi_buffer_pool_caps_is_equal (plugin->sinkpad_buffer_pool, caps))
+  if (sinkpriv->buffer_pool) {
+    if (gst_vaapi_buffer_pool_caps_is_equal (sinkpriv->buffer_pool, caps))
       return TRUE;
-    gst_buffer_pool_set_active (plugin->sinkpad_buffer_pool, FALSE);
-    g_clear_object (&plugin->sinkpad_buffer_pool);
-    g_clear_object (&plugin->sinkpad_allocator);
-    plugin->sinkpad_buffer_size = 0;
+    gst_buffer_pool_set_active (sinkpriv->buffer_pool, FALSE);
+    g_clear_object (&sinkpriv->buffer_pool);
+    g_clear_object (&sinkpriv->allocator);
+    sinkpriv->buffer_size = 0;
   }
 
-  if (!ensure_sinkpad_allocator (plugin, caps, &size))
+  if (!ensure_sinkpad_allocator (plugin, sinkpad, caps, &size))
     return FALSE;
 
   pool =
       gst_vaapi_plugin_base_create_pool (plugin, caps, size,
       BUFFER_POOL_SINK_MIN_BUFFERS, 0,
-      GST_VAAPI_VIDEO_BUFFER_POOL_OPTION_VIDEO_META, plugin->sinkpad_allocator);
+      GST_VAAPI_VIDEO_BUFFER_POOL_OPTION_VIDEO_META, sinkpriv->allocator);
   if (!pool)
     return FALSE;
 
-  plugin->sinkpad_buffer_pool = pool;
-  plugin->sinkpad_buffer_size = size;
+  sinkpriv->buffer_pool = pool;
+  sinkpriv->buffer_size = size;
   return TRUE;
 }
 
+static gboolean
+_set_srcpad_caps (GstVaapiPluginBase * plugin, GstPad * srcpad, GstCaps * caps)
+{
+  GstVaapiPadPrivate *srcpriv = NULL;
+
+  if (caps) {
+    g_assert (srcpad);
+    srcpriv = GST_VAAPI_PAD_PRIVATE (srcpad);
+    g_assert (srcpriv);
+
+    if (caps != srcpriv->caps) {
+      if (!gst_video_info_from_caps (&srcpriv->info, caps))
+        return FALSE;
+      if (srcpriv->buffer_pool
+          && !gst_vaapi_buffer_pool_caps_is_equal (srcpriv->buffer_pool,
+              caps)) {
+        gst_buffer_pool_set_active (srcpriv->buffer_pool, FALSE);
+        g_clear_object (&srcpriv->buffer_pool);
+        g_clear_object (&srcpriv->allocator);
+        plugin_reset_texture_map (plugin);
+      }
+      gst_caps_replace (&srcpriv->caps, caps);
+    }
+  }
+
+  return TRUE;
+}
+
+static gboolean
+_set_sinkpad_caps (GstVaapiPluginBase * plugin, GstPad * sinkpad,
+    GstCaps * caps)
+{
+  GstVaapiPadPrivate *sinkpriv = NULL;
+
+  if (caps) {
+    g_assert (sinkpad);
+    sinkpriv = GST_VAAPI_PAD_PRIVATE (sinkpad);
+    g_assert (sinkpriv);
+
+    if (caps != sinkpriv->caps) {
+      if (!gst_video_info_from_caps (&sinkpriv->info, caps))
+        return FALSE;
+      gst_caps_replace (&sinkpriv->caps, caps);
+      sinkpriv->caps_is_raw = !gst_caps_has_vaapi_surface (caps);
+    }
+
+    if (!ensure_sinkpad_buffer_pool (plugin, sinkpad))
+      return FALSE;
+  }
+
+  return TRUE;
+}
+
+/**
+ * gst_vaapi_plugin_base_pad_set_caps:
+ * @plugin: a #GstVaapiPluginBase
+ * @sinkpad: the sink pad to set @incaps on
+ * @incaps: the sink pad (input) caps
+ * @srcpad: the src pad to set @outcaps on
+ * @outcaps: the src pad (output) caps
+ *
+ * Notifies the base plugin object of the new input and output caps,
+ * obtained from the subclass on the requested pads.
+ *
+ * Returns: %TRUE if the update of caps was successful, %FALSE otherwise.
+ */
+gboolean
+gst_vaapi_plugin_base_pad_set_caps (GstVaapiPluginBase * plugin,
+    GstPad * sinkpad, GstCaps * incaps, GstPad * srcpad, GstCaps * outcaps)
+{
+  return _set_sinkpad_caps (plugin, sinkpad, incaps)
+      && _set_srcpad_caps (plugin, srcpad, outcaps);
+}
+
 /**
  * gst_vaapi_plugin_base_set_caps:
  * @plugin: a #GstVaapiPluginBase
@@ -738,7 +866,7 @@ ensure_sinkpad_buffer_pool (GstVaapiPluginBase * plugin, GstCaps * caps)
  * @outcaps: the src pad (output) caps
  *
  * Notifies the base plugin object of the new input and output caps,
- * obtained from the subclass.
+ * obtained from the subclass, on the base plugin static pads.
  *
  * Returns: %TRUE if the update of caps was successful, %FALSE otherwise.
  */
@@ -746,45 +874,26 @@ gboolean
 gst_vaapi_plugin_base_set_caps (GstVaapiPluginBase * plugin, GstCaps * incaps,
     GstCaps * outcaps)
 {
-  if (incaps && incaps != plugin->sinkpad_caps) {
-    if (!gst_video_info_from_caps (&plugin->sinkpad_info, incaps))
-      return FALSE;
-    gst_caps_replace (&plugin->sinkpad_caps, incaps);
-    plugin->sinkpad_caps_is_raw = !gst_caps_has_vaapi_surface (incaps);
-  }
-
-  if (outcaps && outcaps != plugin->srcpad_caps) {
-    if (!gst_video_info_from_caps (&plugin->srcpad_info, outcaps))
-      return FALSE;
-    if (plugin->srcpad_buffer_pool
-        && !gst_vaapi_buffer_pool_caps_is_equal (plugin->srcpad_buffer_pool,
-            outcaps)) {
-      gst_buffer_pool_set_active (plugin->srcpad_buffer_pool, FALSE);
-      g_clear_object (&plugin->srcpad_buffer_pool);
-      g_clear_object (&plugin->srcpad_allocator);
-      plugin_reset_texture_map (plugin);
-    }
-    gst_caps_replace (&plugin->srcpad_caps, outcaps);
-  }
-
-  if (!ensure_sinkpad_buffer_pool (plugin, plugin->sinkpad_caps))
-    return FALSE;
-  return TRUE;
+  return gst_vaapi_plugin_base_pad_set_caps (plugin, plugin->sinkpad, incaps,
+      plugin->srcpad, outcaps);
 }
 
 /**
- * gst_vaapi_plugin_base_propose_allocation:
+ * gst_vaapi_plugin_base_pad_propose_allocation:
  * @plugin: a #GstVaapiPluginBase
+ * @sinkpad: the sinkpad to configure the allocation query on
  * @query: the allocation query to configure
  *
- * Proposes allocation parameters to the upstream elements.
+ * Proposes allocation parameters to the upstream elements on the requested
+ * sinkpad.
  *
  * Returns: %TRUE if successful, %FALSE otherwise.
  */
 gboolean
-gst_vaapi_plugin_base_propose_allocation (GstVaapiPluginBase * plugin,
-    GstQuery * query)
+gst_vaapi_plugin_base_pad_propose_allocation (GstVaapiPluginBase * plugin,
+    GstPad * sinkpad, GstQuery * query)
 {
+  GstVaapiPadPrivate *sinkpriv = GST_VAAPI_PAD_PRIVATE (sinkpad);
   GstCaps *caps = NULL;
   GstBufferPool *pool = NULL;
   gboolean need_pool;
@@ -794,14 +903,13 @@ gst_vaapi_plugin_base_propose_allocation (GstVaapiPluginBase * plugin,
   if (!caps)
     goto error_no_caps;
 
-  if (!ensure_sinkpad_allocator (plugin, caps, &size))
+  if (!ensure_sinkpad_allocator (plugin, sinkpad, caps, &size))
     return FALSE;
 
   if (need_pool) {
     pool = gst_vaapi_plugin_base_create_pool (plugin, caps, size,
         BUFFER_POOL_SINK_MIN_BUFFERS, 0,
-        GST_VAAPI_VIDEO_BUFFER_POOL_OPTION_VIDEO_META,
-        plugin->sinkpad_allocator);
+        GST_VAAPI_VIDEO_BUFFER_POOL_OPTION_VIDEO_META, sinkpriv->allocator);
     if (!pool)
       return FALSE;
   }
@@ -819,7 +927,7 @@ gst_vaapi_plugin_base_propose_allocation (GstVaapiPluginBase * plugin,
     gst_query_add_allocation_param (query, allocator, NULL);
     gst_object_unref (allocator);
   }
-  gst_query_add_allocation_param (query, plugin->sinkpad_allocator, NULL);
+  gst_query_add_allocation_param (query, sinkpriv->allocator, NULL);
 
   gst_query_add_allocation_pool (query, pool, size,
       BUFFER_POOL_SINK_MIN_BUFFERS, 0);
@@ -838,6 +946,24 @@ error_no_caps:
   }
 }
 
+/**
+ * gst_vaapi_plugin_base_propose_allocation:
+ * @plugin: a #GstVaapiPluginBase
+ * @query: the allocation query to configure
+ *
+ * Proposes allocation parameters to the upstream elements on the base plugin
+ * static sinkpad.
+ *
+ * Returns: %TRUE if successful, %FALSE otherwise.
+ */
+gboolean
+gst_vaapi_plugin_base_propose_allocation (GstVaapiPluginBase * plugin,
+    GstQuery * query)
+{
+  return gst_vaapi_plugin_base_pad_propose_allocation (plugin, plugin->sinkpad,
+      query);
+}
+
 /**
  * gst_vaapi_plugin_base_decide_allocation:
  * @plugin: a #GstVaapiPluginBase
@@ -845,7 +971,8 @@ error_no_caps:
  * @feature: the desired #GstVaapiCapsFeature, or zero to find the
  *   preferred one
  *
- * Decides allocation parameters for the downstream elements.
+ * Decides allocation parameters for the downstream elements on the base
+ * plugin static srcpad.
  *
  * Returns: %TRUE if successful, %FALSE otherwise.
  */
@@ -853,6 +980,7 @@ gboolean
 gst_vaapi_plugin_base_decide_allocation (GstVaapiPluginBase * plugin,
     GstQuery * query)
 {
+  GstVaapiPadPrivate *srcpriv = GST_VAAPI_PAD_PRIVATE (plugin->srcpad);
   GstCaps *caps = NULL;
   GstBufferPool *pool;
   GstVideoInfo vi;
@@ -922,10 +1050,10 @@ gst_vaapi_plugin_base_decide_allocation (GstVaapiPluginBase * plugin,
      * new buffer */
     if (i == 0
         && g_strcmp0 (allocator->mem_type, GST_VAAPI_VIDEO_MEMORY_NAME) != 0) {
-      if (plugin->other_srcpad_allocator)
-        gst_object_unref (plugin->other_srcpad_allocator);
-      plugin->other_srcpad_allocator = allocator;
-      plugin->other_allocator_params = params;
+      if (srcpriv->other_allocator)
+        gst_object_unref (srcpriv->other_allocator);
+      srcpriv->other_allocator = allocator;
+      srcpriv->other_allocator_params = params;
       continue;
     }
 
@@ -933,9 +1061,9 @@ gst_vaapi_plugin_base_decide_allocation (GstVaapiPluginBase * plugin,
       GST_DEBUG_OBJECT (plugin, "found vaapi allocator in query %"
           GST_PTR_FORMAT, allocator);
       index_allocator = i;
-      if (plugin->srcpad_allocator)
-        gst_object_unref (plugin->srcpad_allocator);
-      plugin->srcpad_allocator = allocator;
+      if (srcpriv->allocator)
+        gst_object_unref (srcpriv->allocator);
+      srcpriv->allocator = allocator;
       break;
     }
     gst_object_unref (allocator);
@@ -967,12 +1095,12 @@ gst_vaapi_plugin_base_decide_allocation (GstVaapiPluginBase * plugin,
   }
 
   if (!pool) {
-    if (!ensure_srcpad_allocator (plugin, &vi, caps))
+    if (!ensure_srcpad_allocator (plugin, plugin->srcpad, &vi, caps))
       goto error;
     size = GST_VIDEO_INFO_SIZE (&vi);   /* size might be updated by
                                          * allocator */
     pool = gst_vaapi_plugin_base_create_pool (plugin, caps, size, min, max,
-        pool_options, plugin->srcpad_allocator);
+        pool_options, srcpriv->allocator);
     if (!pool)
       goto error;
   }
@@ -983,19 +1111,19 @@ gst_vaapi_plugin_base_decide_allocation (GstVaapiPluginBase * plugin,
     gst_query_add_allocation_pool (query, pool, size, min, max);
 
   /* allocator might be updated by ensure_srcpad_allocator() */
-  if (plugin->srcpad_allocator) {
+  if (srcpriv->allocator) {
     if (index_allocator > 0) {
       gst_query_set_nth_allocation_param (query, index_allocator,
-          plugin->srcpad_allocator, NULL);
+          srcpriv->allocator, NULL);
     } else {
       GST_DEBUG_OBJECT (plugin, "adding allocator in query %" GST_PTR_FORMAT,
-          plugin->srcpad_allocator);
-      gst_query_add_allocation_param (query, plugin->srcpad_allocator, NULL);
+          srcpriv->allocator);
+      gst_query_add_allocation_param (query, srcpriv->allocator, NULL);
     }
   }
 
-  g_clear_object (&plugin->srcpad_buffer_pool);
-  plugin->srcpad_buffer_pool = pool;
+  g_clear_object (&srcpriv->buffer_pool);
+  srcpriv->buffer_pool = pool;
 
   /* if downstream doesn't support GstVideoMeta, and the negotiated
    * caps are raw video, and the used allocator is the VA-API one, we
@@ -1029,12 +1157,13 @@ error:
 }
 
 /**
- * gst_vaapi_plugin_base_get_input_buffer:
+ * gst_vaapi_plugin_base_pad_get_input_buffer:
  * @plugin: a #GstVaapiPluginBase
+ * @sinkpad: the sink pad to obtain input buffer on
  * @inbuf: the sink pad (input) buffer
  * @outbuf_ptr: the pointer to location to the VA surface backed buffer
  *
- * Acquires the sink pad (input) buffer as a VA surface backed
+ * Acquires the static sink pad (input) buffer as a VA surface backed
  * buffer. This is mostly useful for raw YUV buffers, as source
  * buffers that are already backed as a VA surface are passed
  * verbatim.
@@ -1042,9 +1171,10 @@ error:
  * Returns: #GST_FLOW_OK if the buffer could be acquired
  */
 GstFlowReturn
-gst_vaapi_plugin_base_get_input_buffer (GstVaapiPluginBase * plugin,
-    GstBuffer * inbuf, GstBuffer ** outbuf_ptr)
+gst_vaapi_plugin_base_pad_get_input_buffer (GstVaapiPluginBase * plugin,
+    GstPad * sinkpad, GstBuffer * inbuf, GstBuffer ** outbuf_ptr)
 {
+  GstVaapiPadPrivate *sinkpriv = GST_VAAPI_PAD_PRIVATE (sinkpad);
   GstVaapiVideoMeta *meta;
   GstBuffer *outbuf;
   GstVideoFrame src_frame, out_frame;
@@ -1059,33 +1189,31 @@ gst_vaapi_plugin_base_get_input_buffer (GstVaapiPluginBase * plugin,
     return GST_FLOW_OK;
   }
 
-  if (!plugin->sinkpad_caps_is_raw)
+  if (!sinkpriv->caps_is_raw)
     goto error_invalid_buffer;
 
-  if (!plugin->sinkpad_buffer_pool)
+  if (!sinkpriv->buffer_pool)
     goto error_no_pool;
 
-  if (!gst_buffer_pool_is_active (plugin->sinkpad_buffer_pool) &&
-      !gst_buffer_pool_set_active (plugin->sinkpad_buffer_pool, TRUE))
+  if (!gst_buffer_pool_is_active (sinkpriv->buffer_pool) &&
+      !gst_buffer_pool_set_active (sinkpriv->buffer_pool, TRUE))
     goto error_active_pool;
 
   outbuf = NULL;
-  if (gst_buffer_pool_acquire_buffer (plugin->sinkpad_buffer_pool,
+  if (gst_buffer_pool_acquire_buffer (sinkpriv->buffer_pool,
           &outbuf, NULL) != GST_FLOW_OK)
     goto error_create_buffer;
 
   if (is_dma_buffer (inbuf)) {
-    if (!plugin_bind_dma_to_vaapi_buffer (plugin, inbuf, outbuf))
+    if (!plugin_bind_dma_to_vaapi_buffer (plugin, sinkpad, inbuf, outbuf))
       goto error_bind_dma_buffer;
     goto done;
   }
 
-  if (!gst_video_frame_map (&src_frame, &plugin->sinkpad_info, inbuf,
-          GST_MAP_READ))
+  if (!gst_video_frame_map (&src_frame, &sinkpriv->info, inbuf, GST_MAP_READ))
     goto error_map_src_buffer;
 
-  if (!gst_video_frame_map (&out_frame, &plugin->sinkpad_info, outbuf,
-          GST_MAP_WRITE))
+  if (!gst_video_frame_map (&out_frame, &sinkpriv->info, outbuf, GST_MAP_WRITE))
     goto error_map_dst_buffer;
 
   success = gst_video_frame_copy (&out_frame, &src_frame);
@@ -1155,6 +1283,27 @@ error_copy_buffer:
   }
 }
 
+/**
+ * gst_vaapi_plugin_base_get_input_buffer:
+ * @plugin: a #GstVaapiPluginBase
+ * @inbuf: the sink pad (input) buffer
+ * @outbuf_ptr: the pointer to location to the VA surface backed buffer
+ *
+ * Acquires the static sink pad (input) buffer as a VA surface backed
+ * buffer. This is mostly useful for raw YUV buffers, as source
+ * buffers that are already backed as a VA surface are passed
+ * verbatim.
+ *
+ * Returns: #GST_FLOW_OK if the buffer could be acquired
+ */
+GstFlowReturn
+gst_vaapi_plugin_base_get_input_buffer (GstVaapiPluginBase * plugin,
+    GstBuffer * inbuf, GstBuffer ** outbuf_ptr)
+{
+  return gst_vaapi_plugin_base_pad_get_input_buffer (plugin, plugin->sinkpad,
+      inbuf, outbuf_ptr);
+}
+
 /**
  * gst_vaapi_plugin_base_set_gl_context:
  * @plugin: a #GstVaapiPluginBase
@@ -1255,14 +1404,19 @@ no_valid_gl_display:
 
 static GArray *
 extract_allowed_surface_formats (GstVaapiDisplay * display,
-    GArray * img_formats, GstVideoFormat specified_format,
-    GstPadDirection direction)
+    GArray * img_formats)
 {
   guint i;
   GArray *out_formats;
+  guint surface_chroma = 0;
   GstVaapiSurface *surface = NULL;
 
-  g_assert (direction == GST_PAD_SRC || direction == GST_PAD_SINK);
+  /* let's assume all the formats are possible to download */
+  if (gst_vaapi_display_has_driver_quirks (display,
+          GST_VAAPI_DRIVER_QUIRK_NO_CHECK_SURFACE_PUT_IMAGE)) {
+    out_formats = g_array_ref (img_formats);
+    goto bail;
+  }
 
   out_formats =
       g_array_sized_new (FALSE, FALSE, sizeof (GstVideoFormat),
@@ -1275,49 +1429,34 @@ extract_allowed_surface_formats (GstVaapiDisplay * display,
         g_array_index (img_formats, GstVideoFormat, i);
     GstVaapiImage *image;
     GstVideoInfo vi;
-    GstVideoFormat surface_format;
-    gboolean res;
+    guint img_chroma;
 
     if (img_format == GST_VIDEO_FORMAT_UNKNOWN)
       continue;
 
-    surface_format =
-        (specified_format == GST_VIDEO_FORMAT_UNKNOWN) ?
-        img_format : specified_format;
-    if (!surface) {
-      gst_video_info_set_format (&vi, surface_format, 64, 64);
+    img_chroma = gst_vaapi_video_format_get_chroma_type (img_format);
+    if (img_chroma != surface_chroma) {
+      if (surface)
+        gst_vaapi_surface_unref (surface);
+      gst_video_info_set_format (&vi, img_format, 64, 64);
       surface = gst_vaapi_surface_new_full (display, &vi, 0);
       if (!surface)
         continue;
+      surface_chroma = img_chroma;
     }
 
     image = gst_vaapi_image_new (display, img_format, 64, 64);
-    if (!image) {
-      /* Just reuse the surface if the format is specified */
-      if (specified_format == GST_VIDEO_FORMAT_UNKNOWN)
-        gst_vaapi_object_replace (&surface, NULL);
-
+    if (!image)
       continue;
-    }
-
-    res = FALSE;
-    if (direction == GST_PAD_SRC) {
-      res = gst_vaapi_surface_get_image (surface, image);
-    } else {
-      res = gst_vaapi_surface_put_image (surface, image);
-    }
-    if (res)
+    if (gst_vaapi_surface_put_image (surface, image))
       g_array_append_val (out_formats, img_format);
-
-    gst_vaapi_object_unref (image);
-    /* Just reuse the surface if the format is specified */
-    if (specified_format == GST_VIDEO_FORMAT_UNKNOWN)
-      gst_vaapi_object_replace (&surface, NULL);
+    gst_vaapi_image_unref (image);
   }
 
   if (surface)
-    gst_vaapi_object_unref (surface);
+    gst_vaapi_surface_unref (surface);
 
+bail:
   if (out_formats->len == 0) {
     g_array_unref (out_formats);
     return NULL;
@@ -1326,8 +1465,7 @@ extract_allowed_surface_formats (GstVaapiDisplay * display,
 }
 
 static gboolean
-ensure_allowed_raw_caps (GstVaapiPluginBase * plugin, GstVideoFormat format,
-    GstPadDirection direction)
+ensure_allowed_raw_caps (GstVaapiPluginBase * plugin)
 {
   GArray *formats, *out_formats;
   GstVaapiDisplay *display;
@@ -1342,8 +1480,7 @@ ensure_allowed_raw_caps (GstVaapiPluginBase * plugin, GstVideoFormat format,
   formats = gst_vaapi_display_get_image_formats (display);
   if (!formats)
     goto bail;
-  out_formats =
-      extract_allowed_surface_formats (display, formats, format, direction);
+  out_formats = extract_allowed_surface_formats (display, formats);
   if (!out_formats)
     goto bail;
   out_caps = gst_vaapi_video_format_new_template_caps_from_list (out_formats);
@@ -1375,25 +1512,7 @@ bail:
 GstCaps *
 gst_vaapi_plugin_base_get_allowed_sinkpad_raw_caps (GstVaapiPluginBase * plugin)
 {
-  if (!ensure_allowed_raw_caps (plugin, GST_VIDEO_FORMAT_UNKNOWN, GST_PAD_SINK))
-    return NULL;
-  return plugin->allowed_raw_caps;
-}
-
-/**
- * gst_vaapi_plugin_base_get_allowed_srcpad_raw_caps:
- * @plugin: a #GstVaapiPluginBase
- * @format: a #GstVideoFormat, the format we need to check
- *
- * Returns the raw #GstCaps allowed by the element.
- *
- * Returns: the allowed raw #GstCaps or %NULL
- **/
-GstCaps *
-gst_vaapi_plugin_base_get_allowed_srcpad_raw_caps (GstVaapiPluginBase *
-    plugin, GstVideoFormat format)
-{
-  if (!ensure_allowed_raw_caps (plugin, format, GST_PAD_SRC))
+  if (!ensure_allowed_raw_caps (plugin))
     return NULL;
   return plugin->allowed_raw_caps;
 }
@@ -1404,7 +1523,7 @@ gst_vaapi_plugin_base_get_allowed_srcpad_raw_caps (GstVaapiPluginBase *
  * @object: the GL context from gst-gl
  *
  * This function will determine if @object supports dmabuf
- * importing.
+ * importing on the base plugin static srcpad.
  *
  * Please note that the context @object should come from downstream.
  **/
@@ -1413,9 +1532,10 @@ gst_vaapi_plugin_base_set_srcpad_can_dmabuf (GstVaapiPluginBase * plugin,
     GstObject * object)
 {
 #if USE_EGL && USE_GST_GL_HELPERS
+  GstVaapiPadPrivate *srcpriv = GST_VAAPI_PAD_PRIVATE (plugin->srcpad);
   GstGLContext *const gl_context = GST_GL_CONTEXT (object);
 
-  plugin->srcpad_can_dmabuf =
+  srcpriv->can_dmabuf =
       (!(gst_gl_context_get_gl_api (gl_context) & GST_GL_API_GLES1)
       && gst_gl_context_check_feature (gl_context,
           "EGL_EXT_image_dma_buf_import"));
@@ -1426,7 +1546,7 @@ static void
 _init_performance_debug (void)
 {
 #ifndef GST_DISABLE_GST_DEBUG
-  static volatile gsize _init = 0;
+  static gsize _init = 0;
 
   if (g_once_init_enter (&_init)) {
     GST_DEBUG_CATEGORY_GET (CAT_PERFORMANCE, "GST_PERFORMANCE");
@@ -1452,6 +1572,7 @@ gboolean
 gst_vaapi_plugin_copy_va_buffer (GstVaapiPluginBase * plugin,
     GstBuffer * inbuf, GstBuffer * outbuf)
 {
+  GstVaapiPadPrivate *srcpriv = GST_VAAPI_PAD_PRIVATE (plugin->srcpad);
   GstVideoMeta *vmeta;
   GstVideoFrame src_frame, dst_frame;
   gboolean success;
@@ -1467,11 +1588,9 @@ gst_vaapi_plugin_copy_va_buffer (GstVaapiPluginBase * plugin,
   _init_performance_debug ();
   GST_CAT_INFO (CAT_PERFORMANCE, "copying VA buffer to system memory buffer");
 
-  if (!gst_video_frame_map (&src_frame, &plugin->srcpad_info, inbuf,
-          GST_MAP_READ))
+  if (!gst_video_frame_map (&src_frame, &srcpriv->info, inbuf, GST_MAP_READ))
     return FALSE;
-  if (!gst_video_frame_map (&dst_frame, &plugin->srcpad_info, outbuf,
-          GST_MAP_WRITE)) {
+  if (!gst_video_frame_map (&dst_frame, &srcpriv->info, outbuf, GST_MAP_WRITE)) {
     gst_video_frame_unmap (&src_frame);
     return FALSE;
   }
